/* File: MPI_Fault.c
 * 
 * Hatem Nassrat
 * B00393388 - Dalhousie University
 *
 * A Fault tolerance Implementation for an OLAP system
 *
 */

#include <mpi.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "gflib.h"
#include "MPI_Fault.h"

#define IC_CREATE_TAG 5

int main(int argc, char * argv[])
{
    int i, my_rank, p, f;
    char buf_file[50];
    MPI_Comm *my_comm;
    Element* my_records;


    /*MPI INIT*/
    
    MPI_Init(&argc,&argv);
    MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &p);
    f=p;

    /* set up the MPI description of MPI_Record */

    MPI_Address(&record, &baseaddress);
    lena[0] = 2; MPI_Address(&record.key,&loca[0]);
    loca[0] -= baseaddress; typa[0] = MPI_INT;
    lena[1] = 1; MPI_Address(&record.data   ,&loca[1]);
    loca[1] -= baseaddress; typa[1] = MPI_FLOAT;
    MPI_Type_struct(2, lena, loca, typa, &MPI_Record);
    MPI_Type_commit(&MPI_Record);

        /* create intercommunicators and set error handlers */
    if(my_rank==0) //manager process
    {
        my_comm = (MPI_Comm*) malloc ((p-1)*sizeof(MPI_Comm));
        for(i=1;i<p;i++)
        {
            MPI_Intercomm_create(MPI_COMM_SELF, 0, MPI_COMM_WORLD, i, IC_CREATE_TAG, &my_comm[i-1]);
            MPI_Comm_set_errhandler(my_comm[i-1], MPI_ERRORS_RETURN);
        }
    }
    else
    {
        my_comm = (MPI_Comm*)malloc(sizeof(MPI_Comm));
        MPI_Intercomm_create(MPI_COMM_SELF, 0, MPI_COMM_WORLD, 0, IC_CREATE_TAG, my_comm);
        MPI_Comm_set_errhandler(*my_comm, MPI_ERRORS_RETURN);
    }
    /* Each p has a link to manager and vice versa */

    /*END OF: MPI INIT*/



    /* Init data & compute backups */
    if(my_rank != 0)
    {
        my_records = get_my_records(DATA_SIZE,my_rank,p);
        disk_store("/tmp/my_data",my_records,DATA_SIZE);
        sprintf(buf_file, "/tmp/P%d", my_rank);
        rs_encode("/tmp/my_data", p-1-Kft, Kft, buf_file);
        /*for(i=0;i<p-1;i++) 
        {
            sprintf(buf_file, "/tmp/P%d-%04d.rs", my_rank, i);
            MPI_send_file(buf_file,my_comm,0);
        }
        sprintf(buf_file, "/tmp/P%d-info.txt", my_rank, i);
        MPI_send_file(buf_file,my_comm,0);*/
    }
    else{
        //MPI_recv_file(&my_comm[0],0);
    }
    /* End of init and backup computation*/
   
   /* Finalize */
    if(my_rank==0) 
        for (i=1; i<p; i++) MPI_Comm_free(&my_comm[i-1]);
    else MPI_Comm_free(my_comm);
    MPI_Finalize();
    exit(0);
}

Element* get_my_records(int n, int p_rank, int p_size)
{
    int i;
    Element* array = (Element*)malloc(n*sizeof(Element));

    time_t now;
    (void) time(&now);
    srand48((long) now);
   

    for(i=0;i<n;i++)
    {
        array[i].key = i*(p_size-1)+(p_rank-1);
        array[i].data = lrand48()*.99;
    }
    
    //each processor has a set of sorted keys

    return array;
}

void disk_store(char *fname, Element* array, int n)
{
    FILE *f;
    int i;

    f= fopen(fname,"w");
    for(i=0;i<n;i++)
    {
        fprintf(f,"%d,%f\n",array[i].key,array[i].data);
    }
    
    fclose(f);
}

int MPI_send_file(char* filename, MPI_Comm* com, int dest)
{
    FILE *file;
    char buf[BUFSIZ];
    int err;
    file = fopen(filename,"r");
    
    //sending filename
    err=MPI_Send(filename,BUFSIZ,MPI_CHAR,dest,0,*com);
   
    while(fgets(buf, BUFSIZ, file))
    {
        printf("Sending %s\n",buf);
        err=MPI_Send(buf,BUFSIZ,MPI_CHAR,dest,0,*com);
    }
    buf[0]=0;
    err=MPI_Send(buf,BUFSIZ,MPI_CHAR,dest,0,*com);
    fclose(file);
    return 0;
}

int MPI_recv_file(MPI_Comm *com, int source)
{
    FILE *file;
    char buf[BUFSIZ];
    int err;
    MPI_Status status;

    //recv file name
    err= MPI_Recv(buf, BUFSIZ, MPI_CHAR, source, 0, *com, &status);
    //opening file for writing
    file = fopen(buf,"w");

    while(1)
    {
        err= MPI_Recv(buf, BUFSIZ, MPI_CHAR, source, 0, *com, &status);
        if(buf[0]==0) break;
        fprintf(file,"%s",buf);
        printf("Recv %s",buf);
    }

    fclose(file);
    return 0;
}

/*
Adapted from Procedures and Programs for Galois-Field Arithmetic 
and Reed-Solomon Coding. 
Copyright (C) 2003 James S. Plank

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

Please see http://www.cs.utk.edu/~plank/plank/gflib
for instruction on how to use the GFLIB library.

Modified From main method from (rs_encode.c)
Original code by: James S. Plank (cs.utk.edu/~plank)
Modified by Hatem Nassrat, nassrat@cs.dal.ca

rs_encode
-> encodes file @ "filename", using RS codes to "n" parts + "m" checksums 
files named using the defined "stem"

rs_decode
-> decodes "stemed" files using using data from (n,m, and the vend matrix)
from "stem"-info.txt
*/

rs_encode(char* filename, int n, int m, char* stem)
{
  int i, j, *vdm, *inv, *prod, cache_size;
  int rows, cols, blocksize, orig_size;
  int sz, *factors, tmp, factor;
  char **buffer, *buf_file, *block;
  struct stat buf;
  FILE *f;

  rows = n+m;
  cols = n;

  if (stat(filename, &buf) != 0) {
    perror(filename);
    exit(1);
  }

  sz = buf.st_size;
  orig_size = buf.st_size;
  if (sz % (n*sizeof(unit)) != 0) {
    sz += (n*sizeof(unit) - (sz % (n*sizeof(unit))));
  }
  blocksize = sz/n;

  buffer = (char **) malloc(sizeof(char *)*n);
  for (i = 0; i < n; i++) {
    buffer[i] = (char *) malloc(blocksize);
    if (buffer[i] == NULL) {
      perror("Allocating buffer to store the whole file");
      exit(1);
    }
  }

  f = fopen(filename, "r");
  if (f == NULL) { perror(filename); }
  cache_size = orig_size;

  for (i = 0; i < n; i++) {
    if (cache_size < blocksize) memset(buffer[i], 0, blocksize);
    if (cache_size > 0) {
      if (fread(buffer[i], 1, (cache_size > blocksize) ? blocksize : cache_size, f) <= 0) {
        fprintf(stderr, "Couldn't read the right bytes into the buffer\n");
        exit(1);
      }
    }
    cache_size -= blocksize;
  }
  fclose(f);

  buf_file = (char *) malloc(sizeof(char)*(strlen(stem)+30));
  if (buf_file == NULL) { perror("malloc - buf_file"); exit(1); }
  block = (char *) malloc(sizeof(char)*blocksize);
  if (block == NULL) { perror("malloc - block"); exit(1); }
  for (i = 0; i < n; i++) {
    sprintf(buf_file, "%s-%04d.rs", stem, i);
    printf("Writing %s ...", buf_file); fflush(stdout);
    f = fopen(buf_file, "w");
    if (f == NULL) { perror(buf_file); exit(1); }
    fwrite(buffer[i], 1, blocksize, f);
    fclose(f);
    printf(" Done\n");
  }

  factors = (int *) malloc(sizeof(int)*n);
  if (factors == NULL) { perror("malloc - factors"); exit(1); }

  for (i = 0; i < n; i++) factors[i] = 1;
  
  vdm = gf_make_dispersal_matrix(rows, cols);

  for (i = cols; i < rows; i++) {
    sprintf(buf_file, "%s-%04d.rs", stem, i);
    printf("Calculating  %s ...", buf_file); fflush(stdout);
    memset(block, 0, blocksize); 
    for (j = 0; j < cols; j++) {
      tmp = vdm[i*cols+j]; 
      if (tmp != 0) {
        factor = gf_single_divide(tmp, factors[j]);
/*        printf("M[%02d,%02d] = %3d.  Factors[%02d] = %3d.  Factor = %3d.\n",
                i, j, tmp, j, factors[j], factor); */
        factors[j] = tmp;
/*        printf("     Block %2d Bef: %3d.  ", j, buffer[j][0]); */
        gf_mult_region(buffer[j], blocksize, factor);
/*        printf("Block %2d Aft: %3d.  ", j, buffer[j][0]); */
/*         printf("Block %2d Bef: %3d.  ", i, block[0]); */
        gf_add_parity(buffer[j], block, blocksize);
       /*  printf("Block %2d Aft: %3d.\n", i, block[0]); */
      }
    }
    printf(" writing  ...", buf_file); fflush(stdout);
    f = fopen(buf_file, "w");
    if (f == NULL) { perror(buf_file); exit(1); }
    fwrite(block, 1, blocksize, f);
    printf(" Done\n");
    fclose(f);
  }

  sprintf(buf_file, "%s-info.txt", stem, i);
  f = fopen(buf_file, "w");
  if (f == NULL) { perror(buf_file); exit(1); }
  fprintf(f, "%d\n", orig_size);
  fprintf(f, "%d\n", sz);
  fprintf(f, "%d\n", blocksize);
  fprintf(f, "%d\n", n);
  fprintf(f, "%d\n", m);
  gf_write_matrix(f, vdm, rows, cols);
}
